"""Basic math operations for Unix pipes."""


from __future__ import division

from collections import Counter
import itertools as it
import math
import operator as op
import sys

import click


if sys.version_info.major == 2:  # pragma: no cover
    map = it.imap
    filter = it.ifilter
else:  # pragma: no cover
    from functools import reduce


__version__ = '1.0'
__author__ = 'Kevin Wurster'
__email__ = 'wursterk@gmail.com'
__source__ = 'https://github.com/geowurster/pcalc'
__license__ = '''
New BSD License

Copyright (c) 2016, Kevin D. Wurster
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* The names of pcalc or its contributors may not be used to endorse or
  promote products derived from this software without specific prior written
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''


def _values():
    """Read values from ``stdin`` and cast to ``float``."""
    stream = click.get_text_stream('stdin')
    stream = filter(op.methodcaller('strip'), stream)
    stream = filter(None, stream)
    stream = map(float, stream)

    try:
        first = next(stream)
    except StopIteration:
        raise click.ClickException("stdin didn't get any data.")

    return it.chain([first], stream)


def _cb_fmod(ctx, param, value):

    """Click callback to translate the ``--fmod`` flag directly to a modulo
    function.
    """

    if value:
        func = math.fmod
    else:
        func = op.mod

    return lambda a, b: int(func(a, b))


constant_arg = click.argument('constant', type=click.FLOAT, required=True)
fmod_opt = click.option(
    '--fmod', 'mod_func', is_flag=True, callback=_cb_fmod,
    help="Use Python's 'math.fmod()' function, which is better suited for "
         "floats, instead of the modulo operator.  Causes output to be a "
         "float instead of integer.")


@click.group()
def cli():

    """Basic math operations for Unix pipes.

    When working with a negative positional argument: '$ pcalc mul -- -1'

    All commands read from 'stdin' and write to 'stdout'.  Most commands
    stream but a few (like median) hold all values in memory.  Empty or all
    whitespace lines are skipped.

    Some commands (typically prefixed with 'r') reduce all input values to a
    single value.  For instance, '$ pcalc add 3' adds 3 to all input values,
    but '$ pcalc radd' adds all the values together like:

    \b
        output = 0
        for v in values:
            output = output + v

    For the most part it doesn't matter, but this tool is implemented in
    Python with floating point division enabled when running in Python 2.
    """


@cli.command(name='sum')
def sum_():

    """Compute sum."""

    click.echo(reduce(op.add, _values()))


@cli.command(name='round')
@click.argument('precision', type=click.INT, required=True)
def round_(precision):

    """Round values.

    Precision 0 also casts to 'int'.
    """

    if precision == 0:
        def func(x, _):
            return int(round(x, 0))
    else:
        func = round

    for v in _values():
        click.echo(func(v, precision))


@cli.command()
def ceil():

    """Ceiling values."""

    for v in _values():
        click.echo(int(math.ceil(v)))  # Value not cast to int on Python 2


@cli.command()
def floor():

    """Floor values."""

    for v in _values():
        click.echo(int(math.floor(v)))  # Value not cast to int on Python 2


@cli.command()
@click.argument('denominator', type=click.FLOAT, required=True)
@fmod_opt
def mod(denominator, mod_func):

    """Modulo values by a single divisor.

    Output is dictated by the '--fmod' flag.
    """

    for v in _values():
        click.echo(mod_func(v, denominator))


@cli.command()
@fmod_opt
def rmod(mod_func):

    """Reduce by modulo.

    Output is dictated by the '--fmod' flag.
    """

    click.echo(reduce(mod_func, _values()))


@cli.command()
@constant_arg
def add(constant):

    """Add a constant to values."""

    for v in _values():
        click.echo(v + constant)


@cli.command()
@constant_arg
def sub(constant):

    """Subtract a constant from values."""

    for v in _values():
        click.echo(v - constant)


@cli.command()
@constant_arg
def mul(constant):

    """Multiply values by a constant."""

    for v in _values():
        click.echo(v * constant)


@cli.command()
@constant_arg
def div(constant):

    """Divide values by a constant.

    Floating point division.
    """

    for v in _values():
        click.echo(v / constant)


@cli.command()
def radd():

    """Reduce by addition."""

    click.echo(reduce(op.add, _values()))


@cli.command()
def rsub():
    """Reduce by subtraction."""

    click.echo(reduce(op.sub, _values()))


@cli.command()
def rmul():

    """Reduce by multiplication."""

    click.echo(reduce(op.mul, _values()))


@cli.command()
def rdiv():

    """Reduce by division."""

    click.echo(reduce(op.truediv, _values()))


@cli.command(name='abs')
def abs_():

    """Compute absolute value."""

    for v in _values():
        click.echo(abs(v))


@cli.command()
def median():

    """Compute median."""

    values = sorted(_values())
    if len(values) % 2:
        click.echo(values[int((len(values) - 1) / 2)])
    else:
        stop = int(len(values) / 2) + 1
        start = stop - 2
        middle = values[start:stop]
        click.echo(sum(middle) / 2)


@cli.command()
def mean():

    """Compute mean."""

    values = tuple(_values())
    click.echo(sum(values) / len(values))


@cli.command()
def mode():

    """Compute mode.

    Formatting multiple modes is a little ambiguous in the context of pcalc,
    so this condition triggers an error."""

    count = Counter(_values())

    # If the two most common elements have the same count then there are at
    # least 2 modes.
    if len(count) > 1 and len({c[-1] for c in count.most_common(2)}) == 1:
        raise click.ClickException("Multiple mode's - unsure how to format.")
    else:
        click.echo(count.most_common(1)[0][0])


@cli.command(name='pow')
@constant_arg
def pow_(constant):

    """Exponentiation of values by a constant."""

    for v in _values():
        click.echo(pow(v, constant))



"""Tests for ``pcalc``."""


import os

from click.testing import CliRunner
import pytest

import pcalc


@pytest.fixture('function')
def invoke():
    """Return a function that pins ``click.testing.CliRunner().invoke()`` to
    ``pcalc.cli``.

    Returns
    -------
    function
    """
    
    def _invoke(cmd, input, args=None):

        """Calls ``click.testing.CliRunner().invoke()`` with ``pcalc.cli``,
        the given parameters, and returns the result.

        Parameter
        ---------
        cmd : str
            Subcommand name.
        input : list
            Values to convert to a block of text and pass to ``stdin``.
        args : list or None, optional
            Additional arguments for the subcommand.

        Returns
        -------
        click.testing.Result
        """

        # Convert float values to text block
        input = os.linesep.join(list(map(str, input)))

        if args is None:
            args = []
        else:
            args = list(args)

        return CliRunner().invoke(pcalc.cli, [cmd] + args, input=input)
    
    return _invoke


@pytest.mark.parametrize("cmd,input,expected", [
    ['mean', (0, 10), 5.0],
    ['median', (1, 2, 3), 2.0],
    ['median', (4, 3, 2, 1), 2.5],
    ['radd', (1, 2, 3), 6.0],
    ['rdiv', (100, 10, 2), 5.0],
    ['rmod', (7, -3), -2.0],
    ['rmul', (2, 2, 2), 8.0],
    ['sum', (1, 2, 3), 6.0],
    ['rsub', (1, 2, 3), -4]
])
def test_reducers(invoke, cmd, input, expected):

    """Commands that reduce input values to a single output value."""

    result = invoke(cmd, input)
    assert result.exit_code == 0
    assert float(result.output) == expected


@pytest.mark.parametrize("cmd", [
    'mean', 'median', 'mode', 'radd', 'rdiv', 'rmod', 'rmul', 'sum', 'rsub'])
def test_reducers_single_value(invoke, cmd):

    """Reducers receiving a single value should prdduce only that value."""

    result = invoke(cmd, [-10])
    assert result.exit_code == 0
    assert float(result.output) == -10.0


@pytest.mark.parametrize("input,precision,func", [
    [[1.234567, 3.456789], '0', lambda x: int(round(x, 0))],
    [[1.234567, 3.456789], '1', lambda x: round(x, 1)]
])
def test_round(invoke, input, precision, func):

    """Check rounding to 0 and a specific precision."""

    result = invoke('round', input, args=[precision])
    assert result.exit_code == 0

    expected = map(float, result.output.strip().splitlines())
    expected = map(func, expected)
    expected = map(str, expected)
    expected = os.linesep.join(list(expected))
    assert result.output.strip() == expected.strip()


@pytest.mark.parametrize("cmd,input,args,expected", [
    ['sub', [1.23, -4.56, 6], ['3'], ['-1.77', '-7.56', '3.0']],
    ['add', [1.23, -4.56, 6], ['10'], ['11.23', '5.44', '16.0']],
    ['ceil', [-1.23, 4.0001], [], ['-1', '5']],
    ['floor', [-1.23, 4.0001], [], ['-2', '4']],
    ['div', [-1.23, 100, -000.1], ['3.21'], ['-0.38317001', '31.15264001', '-0.03115001']],
    ['mul', [1, 10, 100], ['--', '-1'], ['-1.0', '-10.0', '-100.0']],
    ['abs', ['1', '-2', '-3.456'], [], ['1.0', '2.0', '3.456']],
    ['pow', ['1', '2', '4'], ['2'], ['1.0', '4.0', '16.0']],
    ['mod', ['7', '5'], ['3'], ['1', '2']]
])
def test_streaming(invoke, cmd, input, args, expected):

    """Commands that stream input values, transform, and emit."""

    result = invoke(cmd, input, args)
    assert result.exit_code == 0

    # Division is harder to check
    if cmd == 'div':
        expected = list(map(float, expected))
        actual = list(map(float, result.output.splitlines()))
        for e, a in zip(expected, actual):
            assert round(e - a, 5) <= 1e-05
    else:
        assert result.output.splitlines() == expected


def test_mode(invoke):
    result = invoke('mode', [1, 5, 8, 10, 10])
    assert result.exit_code == 0
    assert result.output.strip() == '10.0'


def test_mode_exception(invoke):
    result = invoke('mode', [1, 1, 2, 2])
    assert result.exit_code != 0
    assert 'multiple' in result.output.lower()


def test_mod_fmod(invoke):

    """The ``--fmod`` flag changes the output."""

    result = invoke('mod', ['7', '5'], ['3', '--fmod'])
    assert result.exit_code == 0
    assert result.output.splitlines() == ['1', '2']


def test_rmod_fmod(invoke):

    """The ``--fmod`` flag changes the output."""

    result = invoke('rmod', [7, -3], ['--fmod'])
    assert result.exit_code == 0
    assert result.output.strip() == '1'


@pytest.mark.parametrize("cmd", pcalc.cli.commands.keys())
def test_commands_no_data(invoke, cmd):

    """Commands that don't get any data should generally exit with an error
    code, however those that have a required positional argument will exit
    with a slightly different code.  This test is approximate but good enough.
    """
    result = invoke(cmd, [])
    assert result.exit_code != '0'
    if "stdin didn't" not in result.output.lower():
        assert 'missing argument' in result.output.lower()



#!/usr/bin/env python


"""Setup script for ``pcalc``."""


import itertools as it
import os

from setuptools import setup


with open('README.rst') as f:
    readme = f.read().strip()


def parse_dunder_line(string):

    """Take a line like:

        "__version__ = '0.0.8'"

    and turn it into a tuple:

        ('__version__', '0.0.8')

    Not very fault tolerant.
    """

    # Split the line and remove outside quotes
    variable, value = (s.strip() for s in string.split('=')[:2])
    value = value[1:-1].strip()
    return variable, value


with open('pcalc.py') as f:
    dunders = dict(map(
        parse_dunder_line,
        filter(lambda l: l.strip().startswith('__'), f.readlines()[:40])))
    version = dunders['__version__']
    author = dunders['__author__']
    email = dunders['__email__']
    source = dunders['__source__']


extras_require = {
    'dev': [
        'pytest>=3',
        'pytest-cov',
        'coveralls',
    ],
}
extras_require['all'] = list(it.chain.from_iterable(extras_require.values()))


setup(
    name='pcalc',
    author=author,
    author_email=email,
    classifiers=[
        'Topic :: Utilities',
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Topic :: Text Processing',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    description="Basic math operations for Unix pipes.",
    include_package_data=True,
    install_requires=['click'],
    extras_require=extras_require,
    keywords='command line CLI math unix pipe',
    license="New BSD",
    long_description=readme,
    py_modules=['pcalc'],
    url=source,
    version=version,
    zip_safe=True,
    entry_points="""
        [console_scripts]
        pcalc=pcalc:cli
    """
)
